/* GRJ, Apr/May 2020: I edited code from the BADTRIP package
 */


package beast.evolution.tree;


import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.TreeMap;

import beast.core.Description;
import beast.core.Input;
import beast.core.Operator;
import beast.core.StateNode;
import beast.core.StateNodeInitialiser;
import beast.evolution.datatype.Sample;

import static beast.evolution.pomocov.TransmissionDistributions.*;


@Description("TreePoMo, representing the transmission tree for PoMo, enriched with samples and transmission events.")
public class TreePoMo extends StateNode {
    final public Input<TreePoMo> m_initial = new Input<>("initial", "treePoMo to start with");
    final public Input<TraitSetPoMo> InputTraits = new Input<>("traits", "all trait information (samples, epi)", new TraitSetPoMo());

    /**
     * state of dirtiness of a node in the tree
     * DIRTY means a property on the node has changed, but not the topology. "property" includes the node height
     *       and that branch length to its parent.
     * FILTHY means the nodes' parent or child has changed.
     */
    public static final int IS_CLEAN = 0, IS_DIRTY = 1, IS_FILTHY = 2;

    /**
     * counters of number of nodes, nodeCount = internalNodeCount + leafNodeCount *
     */
    protected int nodeCount = -1;
    protected int internalNodeCount = -1;
    protected int leafNodeCount = -1;

    protected beast.evolution.tree.NodePoMo root;
    protected beast.evolution.tree.NodePoMo storedRoot;
    protected NodePoMo[] m_nodes = null;
    protected NodePoMo[] m_storedNodes = null;
    protected List<NodePoMo> extraList;
    protected List<NodePoMo> storedExtraList;

    /**
     * Trait set which specifies all information.
     */
    protected TraitSetPoMo traitSet = null;
    protected List<String> hosts;
    String[] taxa;

    private boolean debugFlag = Boolean.valueOf(System.getProperty("badtrip.debug"));
    private int nCallsToCheckTree = 0;



    @Override
    public void initAndValidate() {
        if (m_initial.get() != null && !(this instanceof StateNodeInitialiser)) {
        	throw new RuntimeException("initial-input should be specified for tree that is not a StateNodeInitialiser");
        }
        
        traitSet=InputTraits.get();
        hosts=traitSet.allHosts;
        taxa = new String[traitSet.labels.size()];
        traitSet.labels.toArray(taxa);
        m_nodes=new NodePoMo[hosts.size()];
        m_storedNodes=new NodePoMo[hosts.size()];
        extraList=new ArrayList<NodePoMo>();
        storedExtraList= new ArrayList<>();

        if (traitSet.getUnobservedCasesFlag()) {
            makeInitialTreeWithIndexCase();
        } else {
            makeInitialTree();
        }
        initArrays();
        System.out.println(root.height);
    }



    // GRJ. Makes a star tree with all transmissions from the root (= the index case)
    private void makeInitialTreeWithIndexCase() {
        sanityChecks();
        int rootHostIdx = argmaxStarts();
        double otherMaxH = maxNonrootStart(rootHostIdx);
        String rootHost = hosts.get(rootHostIdx);
        root = new NodePoMo(rootHost, rootHost);
        root.setEnd(traitSet.mapEnd.get(rootHost));
        root.setStart(traitSet.mapStart.get(rootHost));
        root.isExtra=false;
        root.setHeight(otherMaxH + medianDelayTime());
        root.setNr(0);
        m_nodes[0] = root;
        int idx = 1;
        for (int i = 0; i < hosts.size(); i++) {
            if (i != rootHostIdx) {
                m_nodes[idx] =  new NodePoMo(hosts.get(i), hosts.get(i));
                m_nodes[idx].setEnd(traitSet.mapEnd.get(hosts.get(i)));
                m_nodes[idx].setStart(traitSet.mapStart.get(hosts.get(i)));
                for (int l = 0; l < traitSet.taxa.length; l++) {
                    if (traitSet.getHostValue(taxa[l]).equals(hosts.get(i))){
                        System.out.println("Adding taxon "+taxa[l]+" to host "+hosts.get(i));
                        Sample s = new Sample(traitSet.taxa[traitSet.map.get(taxa[l])], traitSet.getValue(taxa[l]));
                        m_nodes[idx].addSample(s);
                    }
                }
                assert m_nodes[idx].getSamples().size() >= 1;
                m_nodes[idx].sortSamples();
                m_nodes[idx].isExtra=false;
                m_nodes[idx].setParent(m_nodes[0]);
                m_nodes[0].addChild(m_nodes[idx]);
                double infH = 0.5 * (m_nodes[idx].getStart() + m_nodes[idx].getEnd());
                m_nodes[idx].setHeight(infH);
                double originH = 0.5 * (m_nodes[0].getHeight() + infH);
                m_nodes[idx].setOriginHeight(originH);
                m_nodes[idx].setUnobservedCount(sensibleUnobservedCount(originH, infH));
                m_nodes[idx].setNr(idx);
                idx++;
            }
        }
        checkTree();
    }

    // GRJ. For makeInitialTreeWithFixedIndexCase()
    private int argmaxStarts() {
        double highestStart = Double.MIN_VALUE;
        int idxMax = -1;
        for (int i = 0; i < hosts.size(); i++) {
            double start = traitSet.mapStart.get(hosts.get(i));
            if (start > highestStart) {
                highestStart = start;
                idxMax = i;
            }
        }
        assert idxMax >= 0;
        return idxMax;
    }

    // GRJ. For makeInitialTreeWithFixedIndexCase()
    private double maxNonrootStart(int rootIdx) {
        double highestStart = Double.MIN_VALUE;
        for (int i = 0; i < hosts.size(); i++) {
            if (i != rootIdx) {
                highestStart = Math.max(highestStart, traitSet.mapStart.get(hosts.get(i)));
            }
        }
        assert highestStart > 0.0;
        return highestStart;
    }




    //NICOLA: create an initial meaningful TreePoMo ()
    public void makeInitialTree() {
        double EPSILON = 0.000001;
        // make a caterpillar
    	double highest=Double.MIN_VALUE;
    	double lowest=Double.MAX_VALUE;
    	Integer[] connected=new Integer[hosts.size()];
    	int rootNr=-1;
    	for (int i = 0; i < hosts.size(); i++) {
    		connected[i]=0;
    		NodePoMo newN=new NodePoMo(hosts.get(i),hosts.get(i));
    		newN.setEnd(traitSet.mapEnd.get(hosts.get(i)));
    		newN.setStart(traitSet.mapStart.get(hosts.get(i)));
    		for (int l = 0; l < traitSet.taxa.length; l++) {
    			if (traitSet.getHostValue(taxa[l]).equals(hosts.get(i))){
    				System.out.println("Adding taxon "+taxa[l]+" to host "+hosts.get(i));
    				Sample s = new Sample(traitSet.taxa[traitSet.map.get(taxa[l])], traitSet.getValue(taxa[l]));
    				newN.addSample(s);
    			}
    		}
    		newN.sortSamples();
    		newN.isExtra=false;
    		m_nodes[i]=newN;
    		newN.setNr(i);
    		if (newN.getStart()>highest){
    			highest=newN.getStart();
    			root=newN;
    			rootNr=i;
    		}
    		if (newN.getEnd()<lowest){
    			lowest=newN.getEnd();
    		}
    	}
    	root.setParent(null);
    	root.setHeight(root.getStart() - EPSILON);
    	connected[rootNr]=1;
    	int numUnconnected=hosts.size()-1;
    	NodePoMo currentNode=root;
    	NodePoMo nextNode=null;
    	NodePoMo nextUnconnected=null;
    	double heighestNext=Double.MIN_VALUE;
    	while (numUnconnected>0){
    		for (int i = 0; i < hosts.size(); i++) {
    			if (connected[i]==0){
    				if (m_nodes[i].getStart()>currentNode.getEnd()){
    					numUnconnected-=1;
    					connected[i]=1;
    					m_nodes[i].setParent(currentNode);
    					currentNode.addChild(m_nodes[i]);
    					m_nodes[i].setHeight(m_nodes[i].getStart() - EPSILON);
    					if ((nextNode==null || m_nodes[i].getEnd()<nextNode.getEnd() ) && (m_nodes[i].getEnd()<currentNode.getEnd())){
    						nextNode=m_nodes[i];
    					}
    				}else{
    					if (m_nodes[i].getStart()>heighestNext){
    						heighestNext=m_nodes[i].getStart();
    						nextUnconnected=m_nodes[i];
    					}
    				}
    			}
    		}
    		if(numUnconnected>0 && nextNode==null && nextUnconnected!=null){
    			NodePoMo newN=new NodePoMo("unsampled"+(extraList.size()+1),"unsampled"+(extraList.size()+1));
    			newN.setEnd((nextUnconnected.getStart()-nextUnconnected.getEnd())/2.0);
        		newN.setStart((currentNode.getStart()-currentNode.getEnd())/2.0);
        		newN.setParent(currentNode);
        		newN.setHeight(newN.getStart() - EPSILON);
        		extraList.add(newN);
        		newN.setNr(extraList.indexOf(newN));
        		newN.isExtra=true;
        		currentNode.addChild(newN);
        		nextNode=newN;
    		}
    		currentNode.sortChildren();
    		nextUnconnected=null;
    		heighestNext=Double.MIN_VALUE;
    		currentNode=nextNode;
    		nextNode=null;
    	}
    	leafNodeCount=0;
    	for (int i = 1; i < hosts.size(); i++) {
    		if (m_nodes[i].children.size()==0){
    			leafNodeCount+=1;
    		}
    	}
    	nodeCount=hosts.size()+extraList.size();
        internalNodeCount = nodeCount - leafNodeCount;
    }



    protected void initArrays() {
        listNodes(root, m_nodes, extraList);
        m_storedNodes = new NodePoMo[hosts.size()];
        storedExtraList=new ArrayList<NodePoMo>();
        for (int i =0;i<extraList.size();i++){
        	storedExtraList.add(null);
        }
        final NodePoMo copy = root.copy(storedExtraList);
        listNodes(copy, m_storedNodes, storedExtraList);
    }


    // GRJ. For debugging.
    // GRJ TODO. Can add  private boolean debugFlag = Boolean.valueOf(System.getProperty("badtrip.debug"));
    // and check if debugFlag set
    public void checkTree() {
        nCallsToCheckTree++;
        if (nCallsToCheckTree == 74963) {
            System.err.println("BUG  checkTree() nCallsToCheckTree = " + nCallsToCheckTree);
            for (NodePoMo xxx : getNodesAsArray()) {
                System.err.println(xxx.asText());
            }
        }

        if (getUnobservedCases()) {
            assert getExtraList().size() == 0;
        }
        for (NodePoMo node : m_nodes) {
            for (Sample smp : node.getSamples()) {
                assert smp.getTime() <= node.getStart();
                assert smp.getTime() >= node.getEnd();
            }
            assert node.getHeight() <= node.getStart();
            if (!(node.getHeight() >= node.getEnd())) {
                System.err.println("BUG  checkTree() nCallsToCheckTree = " + nCallsToCheckTree);
            }
            assert node.getHeight() >= node.getEnd();
            if (getUnobservedCases()) {
                for (NodePoMo ch : node.getChildren()) {
                    double g = ch.getOriginHeight();
                    double h = ch.getHeight();
                    assert g >= h;
                    assert g <= node.getStart();
                    assert g >= node.getEnd();
                    if (!(g <= node.getHeight())) {
                        System.err.println("BUG  checkTree() nCallsToCheckTree = " + nCallsToCheckTree);
                        for (NodePoMo xxx : getNodesAsArray()) {
                            System.err.println(xxx.asText());
                        }
                    }
                    assert g <= node.getHeight();
                }
            } else {
                for (NodePoMo ch : node.getChildren()) {
                    double h = ch.getHeight();
                    assert h <= node.getStart();
                    assert h >= node.getEnd();
                    assert h <= node.getHeight();
                }
            }
        }
    }

    public int nCheckTreesDone() {
        return nCallsToCheckTree;
    }



    public TreePoMo() {
    }

    public TreePoMo(final NodePoMo rootNode) {
        setRoot(rootNode);
        initArrays();
    }

    /**
     * Construct a tree from newick string -- will not automatically adjust tips to zero.
     */
    public TreePoMo(final String newick) {
    	System.out.println("Method to read PoMo newick not yet implemented!");
    }


    /**
     * getters and setters
     *
     * @return the number of nodes in the beast.tree
     */
    //@Override
	public int getNodeCount() {
        return hosts.size() + extraList.size();
    }

    //@Override
	public int getInternalNodeCount() {
    	internalNodeCount = root.getInternalNodeCount();
        return internalNodeCount;
    }

    //@Override
	public int getLeafNodeCount() {
        leafNodeCount = root.getLeafNodeCount();
        return leafNodeCount;
    }


    /**
     * @return a list of internal (ancestral) nodes contained in this tree, including the root node
     */
    //@Override
	public List<NodePoMo> getInternalNodes() {
        final ArrayList<NodePoMo> internalNodes = new ArrayList<>();
        for (int i = 0; i < getNodeCount(); i++) {
            final NodePoMo node = getNode(i);
            if (!node.isLeaf()) internalNodes.add(node);
        }
        return internalNodes;
    }


    public void setRoot(final NodePoMo root) {
        this.root = root;
        nodeCount = this.root.getNodeCount();
    }



    //@Override
	public NodePoMo getNode(final int nodeNr) {
        return m_nodes[nodeNr];
        //return getNode(nodeNr, root);
    }

    /**
     * @return an array of taxon names in order of their node numbers.
     * Note that in general no special order of taxa is assumed, just the order
     * assumed in this tree. Consider using tree.m_taxonset.get().asStringList()
     * instead.
     */
    public String [] getTaxaNames() {
        return taxa;
    }



    public boolean getUnobservedCases() {
        return traitSet.unobservedCasesInput.get();
    }

    /**
     * convert tree to array representation *
     */
    void listNodes(final NodePoMo node, final NodePoMo[] nodes, final List<NodePoMo> extraList) {
    	if (!node.isExtra){
    		nodes[node.getNr()] = node;
    	}else{
    		extraList.set(node.getNr(), node);
    	}
        node.m_tree = this;  //(JH) I don't understand this code
        // (JH) why not  node.children, we don't keep it around??
        for (final NodePoMo child : node.getChildren()) {
            listNodes(child, nodes, extraList);
        }
    }


    /**
     * @return list of nodes in array format.
     *         *
     */
    //@Override
	public NodePoMo[] getNodesAsArray() {
        return m_nodes;
    }

	
	public List<NodePoMo> getExtraList() {
        return extraList;
    }

    /**
     * deep copy, returns a completely new tree
     *
     * @return a deep copy of this beast.tree.
     */
    @Override
    public TreePoMo copy() {
        TreePoMo tree = new TreePoMo();
        tree.setID(getID());
        tree.index = index;
        tree.extraList = new ArrayList<NodePoMo>();
        tree.root = root.copy(tree.extraList);
        tree.nodeCount = nodeCount;
        tree.internalNodeCount = internalNodeCount;
        tree.leafNodeCount = leafNodeCount;
        tree.taxa=taxa.clone();
        tree.hosts= new ArrayList<String>();
        for (String s : hosts){
        	tree.hosts.add(s);
        }
        return tree;
    }

    /**
     * copy of all values into existing tree *
     */
    @Override
    public void assignTo(final StateNode other) {
        final TreePoMo tree = (TreePoMo) other;
        final NodePoMo[] nodes = new NodePoMo[m_nodes.length];
        final List<NodePoMo> extraList2 = new ArrayList<NodePoMo>();
        for(int i=0;i<this.extraList.size();i++){
        	extraList2.add(null);
        }
        listNodes(tree.root, nodes, extraList2);
        tree.setID(getID());
        root.assignTo(nodes, extraList2);
        if (!root.isExtra){
        	tree.root = nodes[root.getNr()];
        }else{
        	tree.root = extraList2.get(root.getNr());
        }
        tree.nodeCount = nodeCount;
        tree.internalNodeCount = internalNodeCount;
        tree.leafNodeCount = leafNodeCount;
    }
    

    /**
     * copy of all values from existing tree *
     */
    @Override
    public void assignFrom(final StateNode other) {
        final TreePoMo tree = (TreePoMo) other;
        final NodePoMo[] nodes = new NodePoMo[tree.m_nodes.length];
        final List<NodePoMo> extraList2 = new ArrayList<NodePoMo>();
        for (int i = 0; i < tree.m_nodes.length; i++) {
            nodes[i] = new NodePoMo();
        }
        for (int i = 0; i < tree.extraList.size(); i++) {
        	extraList2.add(new NodePoMo());
        }
        setID(tree.getID());
        if(tree.root.isExtra){
        	root = extraList2.get(tree.root.getNr());
        }else{
        	root = nodes[tree.root.getNr()];
        }
        root.assignFrom(nodes,extraList2,  tree.root);
        root.parent = null;
        root.isExtra=tree.root.isExtra;
        nodeCount = tree.nodeCount;
        internalNodeCount = tree.internalNodeCount;
        leafNodeCount = tree.leafNodeCount;
        initArrays();
    }

    /**
     * as assignFrom, but only copy tree structure *
     */
    // GRJ added originHeight, unobservedCount
    @Override
    public void assignFromFragile(final StateNode other) {
        final TreePoMo tree = (TreePoMo) other;
        if (m_nodes == null) {
            initArrays();
        }
        if (tree.root.isExtra){
        	root = extraList.get(tree.root.getNr());
        }else{
        	root = m_nodes[tree.root.getNr()];
        }
        final NodePoMo[] otherNodes = tree.m_nodes;
        final int rootNr = root.getNr();
        assignFrom(0, rootNr, otherNodes);
        root.isExtra=tree.root.isExtra;
        root.setNr(tree.root.getNr());
        root.start=tree.root.start;
        root.end=tree.root.end;
        root.height = otherNodes[rootNr].height;
        root.originHeight = otherNodes[rootNr].originHeight;
        root.unobservedCount = otherNodes[rootNr].unobservedCount;
        root.parent = null;
        for(int i=0;i<tree.root.samples.size();i++){
        	root.addSample(tree.root.samples.get(i));
        }
        for (int i=0; i<otherNodes[rootNr].children.size();i++){
        	if(otherNodes[rootNr].getChild(i).isExtra){
        		root.addChild(extraList.get(otherNodes[rootNr].getChild(i).getNr()));
        	}else{
        		root.addChild(m_nodes[otherNodes[rootNr].getChild(i).getNr()]);
        	}
        }
        assignFrom(rootNr + 1, m_nodes.length+extraList.size(), otherNodes);
    }


    /**
     * helper to assignFromFragile *
     */
    // GRJ added originHeight, unobservedCount
    private void assignFrom(final int start, final int end, final NodePoMo[] otherNodes) {
        for (int i = start; i < end; i++) {
            NodePoMo sink = m_nodes[i];
            NodePoMo src = otherNodes[i];
            sink.height = src.height;
            sink.originHeight = src.originHeight;
            sink.unobservedCount = src.unobservedCount;
            if (sink.isExtra){
            	sink.parent = extraList.get(src.parent.getNr());
            }else{
            	sink.parent = m_nodes[src.parent.getNr()];
            }
            sink.isExtra=src.isExtra;
            sink.setNr(src.getNr());
            sink.start=src.start;
            sink.end=src.end;
            for(int s=0;s<src.samples.size();s++){
            	sink.addSample(src.samples.get(s));
            }
            for (int c=0; c<src.children.size();c++){
            	if(src.getChild(c).isExtra){
            		sink.addChild(extraList.get(src.getChild(c).getNr()));
            	}else{
            		sink.addChild(m_nodes[src.getChild(c).getNr()]);
            	}
            }
        }
    }


    @Override
	public String toString() {
        return root.toString();
    }


    /**
     * StateNode implementation
     */
    @Override
    public void setEverythingDirty(final boolean isDirty) {
        setSomethingIsDirty(isDirty);
        if (!isDirty) {
            for( NodePoMo n : m_nodes ) {
                n.isDirty = IS_CLEAN;
            }
            for( NodePoMo n : extraList ) {
                n.isDirty = IS_CLEAN;
            }
          //  root.makeAllDirty(IS_CLEAN);
        } else {
            for( NodePoMo n : m_nodes ) {
                n.isDirty = IS_FILTHY;
            }
            for( NodePoMo n : extraList ) {
                n.isDirty = IS_FILTHY;
            }
        //    root.makeAllDirty(IS_FILTHY);
        }
    }

    @Override
    public int scale(final double scale) {
        int s=root.scale(scale);
        checkTree();
        return s;
    }



    /** Loggable interface implementation follows **/

    /**
     * print translate block for NEXUS beast.tree file
     */
    public static void printTranslate(final NodePoMo node, final PrintStream out, final int nodeCount) {
        final List<String> translateLines = new ArrayList<>();
        printTranslate(node, translateLines, nodeCount);
        Collections.sort(translateLines);
        for (final String line : translateLines) {
            out.println(line);
        }
    }

    static public int taxaTranslationOffset = 1;

    /**
     * need this helper so that we can sort list of entries *
     */
    static void printTranslate(NodePoMo node, List<String> translateLines, int nodeCount) {
    	final String nr = (node.getNr() + taxaTranslationOffset) + "";
        String line = "\t\t" + "    ".substring(nr.length()) + nr + " " + node.getID();
        if (node.getNr() < nodeCount) {
            line += ",";
        }
        translateLines.add(line);
        if (!(node.isLeaf())) {
        	for(int i=0;i<node.children.size();i++){
        		 printTranslate(node.getChild(i), translateLines, nodeCount);
        	}
        }
    }

    public static void printTaxa(final NodePoMo node, final PrintStream out, final int nodeCount) {
        final List<String> translateLines = new ArrayList<>();
        printTranslate(node, translateLines, nodeCount);
        Collections.sort(translateLines);
        for (String line : translateLines) {
            line = line.split("\\s+")[2];
            out.println("\t\t\t" + line.replace(',', ' '));
        }
    }
    
    //@Override
	public NodePoMo getRoot() {
        return root;
    }

    @Override
	public void init(PrintStream out) {
        NodePoMo node = getRoot();
        out.println("#NEXUS\n");
        out.println("Begin taxa;");
        //int leaves=getLeafNodeCount();
        int leaves=nodeCount;
        out.println("\tDimensions ntax=" + leaves + ";");
        out.println("\t\tTaxlabels");
        printTaxa(node, out, leaves);
        out.println("\t\t\t;");
        out.println("End;");

        out.println("Begin trees;");
        out.println("\tTranslate");
        printTranslate(node, out, leaves);
        out.print(";");
    }


    @Override
	public void log(long sample, PrintStream out) {
        TreePoMo tree = (TreePoMo) getCurrent();
        out.print("tree STATE_" + sample + " = ");
        // Don't sort, this can confuse CalculationNodes relying on the tree
        final int[] dummy = new int[1];
        final String newick = tree.getRoot().toSortedNewick(dummy);
        out.print(newick);
        out.print(";");
    }

    /**
     * @see beast.core.Loggable *
     */
    @Override
	public void close(PrintStream out) {
        out.print("End;");
    }

    /**
     * reconstruct tree from XML fragment in the form of a DOM node *
     */
    @Override
    public void fromXML(final org.w3c.dom.Node node) {
        System.out.println("function to read from PoMo newick tree not yet implemented!");
        System.exit(0);
        initArrays();
    }

    /**
     * Valuable implementation *
     */
    @Override
	public int getDimension() {
        return getNodeCount();
    }

    //@Override
	public double getArrayValue() {
        return root.height;
    }

    //@Override
	public double getArrayValue(int value, boolean isE) {
		if (isE){
			return extraList.get(value).height;
		}else{
			return m_nodes[value].height;
		}
    }
	
	public double getArrayValue(int value) {
		return m_nodes[value].height;
    }
	

    /**
     * StateNode implementation *
     */
    @Override
    protected void store() {
    	//System.out.println("Storing tree start ");
    	//this should not matter as the number of observed nodes is not modifiable
    	if (m_storedNodes.length != m_nodes.length) {
            final NodePoMo[] tmp = new NodePoMo[m_nodes.length];
            System.arraycopy(m_storedNodes, 0, tmp, 0, m_storedNodes.length - 1);
            if (m_nodes.length > m_storedNodes.length) {
                tmp[m_storedNodes.length - 1] = m_storedNodes[m_storedNodes.length - 1];
                tmp[m_nodes.length - 1] = new NodePoMo();
                tmp[m_nodes.length - 1].setNr(m_nodes.length - 1);
            }
            m_storedNodes = tmp;
        }
    	//fix size of stored extra node list
    	if (extraList.size() > storedExtraList.size()) {
        	for (int i=storedExtraList.size();i<extraList.size();i++){
        		storedExtraList.add(new NodePoMo());
        		storedExtraList.get(i).setNr(i);
        		storedExtraList.get(i).isExtra=true;
        	}
        }else if (extraList.size() < storedExtraList.size()) {
        	for (int i=extraList.size();i<storedExtraList.size();i++){
        		storedExtraList.remove(storedExtraList.size()-1);
        	}
        }

        storeNodes(0, m_nodes.length);
        storeNodesExtra(0, extraList.size());
        if (root.isExtra){
        	storedRoot = storedExtraList.get(root.getNr());
        }else{
        	storedRoot = m_storedNodes[root.getNr()];
        }
        //System.out.println("Storing tree end ");
    }


    /**
     * Stores nodes with index i, for start <= i < end
     * (i.e. including start but not including end)
     *
     * @param start the first index to be stored
     * @param end   nodes are stored up to but not including this index
     */
    private void storeNodes(final int start, final int end) {
        // Use direct members for speed (we are talking 5-7% or more from total time for large trees :)
        for (int i = start; i < end; i++) {
            final NodePoMo sink = m_storedNodes[i];
            final NodePoMo src = m_nodes[i];
            storeNode(sink, src, i);
        }
    }
    
    private void storeNodesExtra(final int start, final int end) {
        // Use direct members for speed (we are talking 5-7% or more from total time for large trees :)
        for (int i = start; i < end; i++) {
            final NodePoMo sink = storedExtraList.get(i);
            final NodePoMo src = extraList.get(i);
            storeNode(sink, src, i);
        }
    }

    // GRJ added originHeight, unobservedCount
    private void storeNode(NodePoMo sink, NodePoMo src, final int i) {
        // Use direct members for speed (we are talking 5-7% or more from total time for large trees :)
        sink.height = src.height;
        sink.originHeight = src.originHeight;
        sink.unobservedCount = src.unobservedCount;
        sink.start = src.start;
        sink.end = src.end;
        sink.labelNr=i;
        sink.isExtra=src.isExtra;
        sink.setName(src.hostName);
        sink.metaDataString = src.metaDataString;
        sink.metaData = new TreeMap<>(src.metaData);
        sink.setID(src.getID());
        sink.samples=new ArrayList<Sample>();
        for(int j=0;j<src.samples.size();j++){
            sink.addSample(src.samples.get(j));
        }

        if ( src.parent != null ) {
            if (src.parent.isExtra){
                sink.parent = storedExtraList.get(src.parent.getNr());
            }else{
                sink.parent = m_storedNodes[src.parent.getNr()];
            }
        } else {
            sink.parent = null;
        }

        final List<NodePoMo> children = sink.children;
        final List<NodePoMo> srcChildren = src.children;

        if( children.size() == srcChildren.size() ) {
            // save some more time by avoiding list clear and add
            for (int k = 0; k < children.size(); ++k) {
                final NodePoMo srcChild = srcChildren.get(k);
                final NodePoMo c;
                if (srcChild.isExtra){
                    c = storedExtraList.get(srcChild.getNr());
                }else{
                    c = m_storedNodes[srcChild.getNr()];
                }
                c.parent = sink;
                children.set(k, c);
            }
        } else {
            children.clear();
            for (final NodePoMo srcChild : srcChildren) {
                final NodePoMo c;
                if (srcChild.isExtra){
                    c = storedExtraList.get(srcChild.getNr());
                }else{
                    c = m_storedNodes[srcChild.getNr()];
                }
                c.parent = sink;
                children.add(c);
            }
        }
    }


    @Override
    public void startEditing(final Operator operator) {
        super.startEditing(operator);
    }

    @Override
    public void restore() {
        // necessary for sampled ancestor trees
        nodeCount = m_storedNodes.length + storedExtraList.size();

        final NodePoMo[] tmp = m_storedNodes;
        m_storedNodes = m_nodes;
        m_nodes = tmp;
        final List<NodePoMo> tmp2 = storedExtraList;
        storedExtraList = extraList;
        extraList = tmp2;
        if (storedRoot.isExtra){
        	root = extraList.get(storedRoot.getNr());
        }else{
        	root = m_nodes[storedRoot.getNr()];
        }
        leafNodeCount = 0;
        for( NodePoMo n : m_nodes ) {
            leafNodeCount += n.isLeaf() ? 1 : 0;
        }
        for( NodePoMo n : extraList ) {
            leafNodeCount += n.isLeaf() ? 1 : 0;
        }
        hasStartedEditing = false;
        for( NodePoMo n : m_nodes ) {
            n.isDirty = Tree.IS_CLEAN;
        }
        for( NodePoMo n : extraList ) {
            n.isDirty = Tree.IS_CLEAN;
        }
    }


    /**
     * Convert age/height to the date time scale given by a trait set,
     * if one exists.  Otherwise just return the unconverted height.
     *
     * @param height
     * @return date specified by height
     */
    public double getDate(final double height) {
    	return traitSet.getDate(height);
    }



} // class Tree